from collections import defaultdict

from django.db.models import Exists, OuterRef
from rest_framework import serializers
from rest_framework.exceptions import PermissionDenied, ValidationError

from arkindex.documents.models import Corpus
from arkindex.process.models import Process, ProcessDatasetSet, ProcessMode, Task
from arkindex.project.mixins import ProcessACLMixin
from arkindex.training.models import DatasetSet
from arkindex.training.serializers import DatasetSerializer
from arkindex.users.models import Role


def _set_id_from_context(serializer_field):
    return serializer_field.context.get("view").kwargs["set"]


def _process_id_from_context(serializer_field):
    return serializer_field.context.get("view").kwargs["process"]


_set_id_from_context.requires_context = True
_process_id_from_context.requires_context = True


class ProcessDatasetSetSerializer(ProcessACLMixin, serializers.ModelSerializer):
    process_id = serializers.HiddenField(
        write_only=True,
        default=_process_id_from_context
    )
    set_id = serializers.HiddenField(
        write_only=True,
        default=_set_id_from_context
    )
    dataset = DatasetSerializer(read_only=True, source="set.dataset")
    set_name = serializers.CharField(read_only=True, source="set.name")

    class Meta:
        model = ProcessDatasetSet
        fields = ("process_id", "set_id", "id", "dataset", "set_name", )
        read_only_fields = ("process_id", "id", "dataset", "set_name", )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get("request"):
            # Do not raise Error in order to create OpenAPI schema
            return
        # Required for the ProcessACLMixin and readable corpora
        self._user = self.context["request"].user

    def validate(self, data):
        request_method = self.context["request"].method
        errors = defaultdict(list)

        # Validate process
        process_qs = (
            Process
            .objects
            # Avoid a stale read when adding a dataset right after creating a process
            .using("default")
            # Required for ACL checks
            .select_related("corpus")
            # Required to check for a process that has already started
            .annotate(has_tasks=Exists(Task.objects.filter(process=OuterRef("pk"))))
        )
        if not self.instance:
            try:
                process = process_qs.get(pk=data["process_id"])
            except Process.DoesNotExist:
                raise ValidationError({"process": [f'Invalid pk "{str(data["process_id"])}" - object does not exist.']})
        else:
            process = self.instance.process
        access = self.process_access_level(process)
        if not access or not (access >= Role.Admin.value):
            raise PermissionDenied(detail="You do not have admin access to this process.")

        # Validate dataset set
        if not self.instance:
            if request_method == "DELETE":
                # Allow deleting ProcessDatasetSets even if the user looses access to the corpus
                set_qs = DatasetSet.objects.all()
            else:
                set_qs = (
                    DatasetSet.objects.filter(dataset__corpus__in=Corpus.objects.readable(self._user))
                    .select_related("dataset__creator").prefetch_related("dataset__sets")
                )
            try:
                dataset_set = set_qs.get(pk=data["set_id"])
            except DatasetSet.DoesNotExist:
                raise ValidationError({"set": [f'Invalid pk "{str(data["set_id"])}" - object does not exist.']})
        else:
            dataset_set = self.instance.set

        if process.mode != ProcessMode.Dataset:
            errors["process"].append('Dataset sets can only be added to or removed from processes of mode "dataset".')

        if process.has_tasks:
            errors["process"].append("Dataset sets cannot be updated on processes that have already started.")

        if self.context["request"].method == "POST" and process.sets.filter(id=dataset_set.id).exists():
            errors["set"].append("This dataset set is already selected in this process.")

        if errors:
            raise ValidationError(errors)

        data["set"] = dataset_set
        return data
