from io import StringIO
from textwrap import dedent

from django.core.management import call_command
from django.core.management.base import CommandError

from arkindex.process.models import Process, ProcessMode, WorkerVersion
from arkindex.project.tests import FixtureTestCase
from arkindex.users.models import User


class TestFakeWorkerRun(FixtureTestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.worker_version = WorkerVersion.objects.get(worker__slug="reco")

    def fake_worker_run(self, args=[]):
        """
        Run the fake_worker_run command, capturing its output
        """
        output = StringIO()
        call_command("fake_worker_run", args + ["--no-color"], stdout=output, stderr=output)
        return output.getvalue().strip()

    def test_create_worker_run(self):
        Process.objects.filter(mode=ProcessMode.Local, creator=self.user).delete()

        output = self.fake_worker_run(["--user", str(self.user.id), "--worker-version", str(self.worker_version.id)])

        process = Process.objects.get(mode=ProcessMode.Local, creator=self.user)
        worker_run = process.worker_runs.get()
        self.assertEqual(worker_run.version, self.worker_version)
        self.assertIsNone(worker_run.configuration_id)
        self.assertIsNone(worker_run.model_version_id)
        self.assertListEqual(worker_run.parents, [])

        self.assertEqual(output, dedent(
            f"""
            Created local process {process.id}
            Created WorkerRun {worker_run.id}
            """
        ).strip())

    def test_existing_process(self):
        process = Process.objects.get(mode=ProcessMode.Local, creator=self.user)
        process.worker_runs.all().delete()

        output = self.fake_worker_run(["--user", str(self.user.id), "--worker-version", str(self.worker_version.id)])

        worker_run = process.worker_runs.get()
        self.assertEqual(worker_run.version, self.worker_version)
        self.assertIsNone(worker_run.configuration_id)
        self.assertIsNone(worker_run.model_version_id)
        self.assertListEqual(worker_run.parents, [])

        self.assertEqual(output, dedent(
            f"""
            Using existing local process {process.id}
            Created WorkerRun {worker_run.id}
            """
        ).strip())

    def test_existing_worker_run(self):
        process = Process.objects.get(mode=ProcessMode.Local, creator=self.user)
        worker_run = process.worker_runs.create(version=self.worker_version, parents=[])
        self.assertEqual(process.worker_runs.count(), 2)

        output = self.fake_worker_run(["--user", str(self.user.id), "--worker-version", str(self.worker_version.id)])

        self.assertEqual(process.worker_runs.count(), 2)

        self.assertEqual(output, dedent(
            f"""
            Using existing local process {process.id}
            A WorkerRun for this worker version already exists: {worker_run.id}
            """
        ).strip())

    def test_unknown_worker_version(self):
        worker_version_id = "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"
        self.assertFalse(WorkerVersion.objects.filter(id=worker_version_id).exists())

        with self.assertRaises(CommandError) as ctx:
            self.fake_worker_run(["--user", str(self.user.id), "--worker-version", str(worker_version_id)])

        self.assertEqual(str(ctx.exception), 'WorkerVersion "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa" does not exist')

    def test_unknown_user(self):
        user_id = 12345
        self.assertFalse(User.objects.filter(id=user_id).exists())

        with self.assertRaises(CommandError) as ctx:
            self.fake_worker_run(["--user", str(user_id), "--worker-version", str(self.worker_version.id)])

        self.assertEqual(str(ctx.exception), 'User "12345" does not exist')

    def test_email(self):
        Process.objects.filter(mode=ProcessMode.Local, creator=self.user).delete()

        output = self.fake_worker_run(["--user", str(self.user.email), "--worker-version", str(self.worker_version.id)])

        process = Process.objects.get(mode=ProcessMode.Local, creator=self.user)
        worker_run = process.worker_runs.get()
        self.assertEqual(worker_run.version, self.worker_version)
        self.assertIsNone(worker_run.configuration_id)
        self.assertIsNone(worker_run.model_version_id)
        self.assertListEqual(worker_run.parents, [])

        self.assertEqual(output, dedent(
            f"""
            Created local process {process.id}
            Created WorkerRun {worker_run.id}
            """
        ).strip())
