import uuid
from datetime import datetime, timezone
from unittest.mock import call, patch

from django.urls import reverse
from rest_framework import status

from arkindex.process.models import (
    FeatureUsage,
    Process,
    ProcessMode,
    Repository,
    WorkerConfiguration,
    WorkerRun,
    WorkerVersion,
    WorkerVersionState,
)
from arkindex.project.tests import FixtureAPITestCase
from arkindex.training.models import Model, ModelVersion, ModelVersionState
from arkindex.users.models import Right, Role, User


class TestUserWorkerRuns(FixtureAPITestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.version_1 = WorkerVersion.objects.get(worker__slug="reco")
        cls.worker_1 = cls.version_1.worker
        cls.custom_version = WorkerVersion.objects.get(worker__slug="custom")
        cls.other_version = WorkerVersion.objects.create(
            worker=cls.custom_version.worker,
            revision=None,
            version=2,
            configuration={
                "name": "bartimaeus",
                "category": "level 14 djinni"
            },
            state=WorkerVersionState.Created,
        )

        cls.model = Model.objects.create(name="Some model", public=False)
        cls.model_version = cls.model.versions.create(hash="bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb", archive_hash="aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa", size=8)

        # Local worker run
        cls.local_process = Process.objects.get(mode=ProcessMode.Local, creator=cls.user)
        cls.local_run = WorkerRun.objects.get(process=cls.local_process)
        # Give execution rights to self.user on the custom worker
        Right.objects.create(user=cls.user, content_object=cls.custom_version.worker, level=Role.Contributor.value)

    # ListUserWorkerRuns

    def test_list_user_runs_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.get(reverse("api:user-worker-run-list"))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_list_user_runs_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.get(reverse("api:user-worker-run-list"))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch("arkindex.documents.export.datetime")
    def test_list_user_runs(self, datetime_mock):
        datetime_mock.now.return_value = datetime(2100, 1, 1, 12, 34, 56, tzinfo=timezone.utc)
        test_local_run = WorkerRun.objects.create(
            process=self.local_process,
            version=self.version_1,
            parents=[],
        )
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.get(reverse("api:user-worker-run-list"))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(response.json()["results"], [{
            "configuration": None,
            "id": str(test_local_run.id),
            "model_version": None,
            "parents": [],
            "process": {
                "activity_state": "disabled",
                "corpus": None,
                "chunks": 1,
                "id": str(self.local_process.id),
                "mode": "local",
                "name": None,
                "state": "unscheduled",
                "use_cache": False,
            },
            "summary": "Worker Recognizer @ version 1",
            "worker_version": {
                "configuration": {
                    "test": 42
                },
                "created": self.version_1.created.isoformat().replace("+00:00", "Z"),
                "docker_image_iid": self.version_1.docker_image_iid,
                "gpu_usage": "disabled",
                "id": str(self.version_1.id),
                "model_usage": FeatureUsage.Disabled.value,
                "revision": None,
                "state": "available",
                "version": 1,
                "worker": {
                    "id": str(self.version_1.worker.id),
                    "name": "Recognizer",
                    "slug": "reco",
                    "type": "recognizer"
                }
            },
            "use_gpu": False
        }, {
            "configuration": None,
            "id": str(self.local_run.id),
            "model_version": None,
            "parents": [],
            "process": {
                "activity_state": "disabled",
                "corpus": None,
                "chunks": 1,
                "id": str(self.local_process.id),
                "mode": "local",
                "name": None,
                "state": "unscheduled",
                "use_cache": False,
            },
            "summary": "Worker Custom worker @ version 1",
            "worker_version": {
                "configuration": {
                    "custom": "value"
                },
                "created": self.custom_version.created.isoformat().replace("+00:00", "Z"),
                "docker_image_iid": None,
                "gpu_usage": "disabled",
                "id": str(self.custom_version.id),
                "model_usage": FeatureUsage.Disabled.value,
                "revision": None,
                "state": "created",
                "version": 1,
                "worker": {
                    "id": str(self.custom_version.worker.id),
                    "name": "Custom worker",
                    "slug": "custom",
                    "type": "custom"
                }
            },
            "use_gpu": False
        }])

    def test_list_user_runs_only_own_runs(self):
        write_user = User.objects.get(email="user2@user.fr")
        test_local_process = Process.objects.create(
            mode=ProcessMode.Local,
            creator=write_user
        )
        test_local_run = WorkerRun.objects.create(
            process=test_local_process,
            version=self.version_1,
            parents=[],
        )
        assert WorkerRun.objects.filter(process__mode=ProcessMode.Local, process__creator=write_user).count() == 1
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.get(reverse("api:user-worker-run-list"))
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.json()["results"]), 1)
        assert test_local_run.id not in [item["id"] for item in response.json()["results"]]

    def test_list_user_runs_no_local_process(self):
        write_user = User.objects.get(email="user2@user.fr")
        assert Process.objects.filter(mode=ProcessMode.Local, creator=write_user).count() == 0
        self.client.force_login(write_user)
        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:user-worker-run-list"))
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()["count"], 0)
        self.assertEqual(response.json()["results"], [])

    # CreateUserWorkerRun

    def test_create_user_run_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.post(reverse("api:user-worker-run-create"))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_create_user_run_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.post(reverse("api:user-worker-run-create"))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_create_user_run(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse("api:user-worker-run-create"),
                data={"worker_version_id": str(self.other_version.id)}
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        created_run = WorkerRun.objects.get(version=self.other_version)
        self.assertDictEqual(response.json(), {
            "id": str(created_run.id),
            "parents": [],
            "configuration": None,
            "model_version": None,
            "worker_version": {
                "configuration": {
                    "name": "bartimaeus",
                    "category": "level 14 djinni"
                },
                "created": self.other_version.created.isoformat().replace("+00:00", "Z"),
                "docker_image_iid": None,
                "gpu_usage": "disabled",
                "id": str(self.other_version.id),
                "model_usage": FeatureUsage.Disabled.value,
                "revision": None,
                "state": "created",
                "version": 2,
                "worker": {
                    "id": str(self.other_version.worker.id),
                    "name": "Custom worker",
                    "slug": "custom",
                    "type": "custom"
                }
            },
            "summary": "Worker Custom worker @ version 2",
            "process": {
                "activity_state": "disabled",
                "corpus": None,
                "chunks": 1,
                "id": str(self.local_process.id),
                "mode": "local",
                "name": None,
                "state": "unscheduled",
                "use_cache": False,
            },
            "use_gpu": False
        })

    def test_create_user_run_no_local_process(self):
        """
        If the user doesn't already have a local process, one is created
        """
        test_user = User.objects.get(email="user2@user.fr")
        Right.objects.create(user=test_user, content_object=self.other_version.worker, level=Role.Contributor.value)
        self.client.force_login(test_user)
        self.assertFalse(Process.objects.filter(mode=ProcessMode.Local, creator=test_user).exists())
        with self.assertNumQueries(10):
            response = self.client.post(
                reverse("api:user-worker-run-create"),
                data={"worker_version_id": str(self.other_version.id)}
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertTrue(Process.objects.filter(mode=ProcessMode.Local, creator=test_user).exists())

    def test_create_user_run_wv_doesnt_exist(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:user-worker-run-create"),
                data={"worker_version_id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"worker_version_id": ["Worker version aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa does not exist."]})

    def test_create_user_run_wv_has_revision(self):
        self.version_1.worker.repository = Repository.objects.create(url="http://nerv.co.jp/eva/")
        self.version_1.worker.save()
        self.version_1.revision = self.version_1.worker.repository.revisions.create(
            hash=str(uuid.uuid4()),
            message="A revision",
            author="Someone",
        )
        self.version_1.version = None
        self.version_1.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:user-worker-run-create"),
                data={"worker_version_id": str(self.version_1.id)}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"worker_version_id": ["The worker version used to create a local worker run must not have a revision set."]})

    def test_create_user_run_worker_has_repository(self):
        self.version_1.worker.repository = Repository.objects.create(url="http://nerv.co.jp/eva/")
        self.version_1.worker.save()
        test_version = self.version_1.worker.versions.create(configuration={}, version=7)
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:user-worker-run-create"),
                data={"worker_version_id": str(test_version.id)}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"worker_version_id": ["The worker used to create a local worker run must not be linked to a repository."]})

    def test_create_user_run_worker_archived(self):
        self.custom_version.worker.archived = datetime.now(timezone.utc)
        self.custom_version.worker.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:user-worker-run-create"),
                data={"worker_version_id": str(self.custom_version.id)}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {"worker_version_id": ["The worker used to create a local worker run must not be archived."]})

    @patch("arkindex.users.utils.get_max_level", return_value=Role.Guest.value)
    def test_create_user_run_worker_not_executable(self, get_max_level_mock):
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:user-worker-run-create"),
                data={"worker_version_id": str(self.other_version.id)}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {"worker_version_id": ["You do not have contributor access to this worker."]})

        self.assertEqual(get_max_level_mock.call_count, 1)
        self.assertEqual(get_max_level_mock.call_args, call(self.user, self.other_version.worker))

    def test_create_user_run_mv_doesnt_exist(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:user-worker-run-create"),
                data={"worker_version_id": str(self.other_version.id), "model_version_id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"model_version_id": ["Model version aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa does not exist."]})

    def test_create_user_run_wc_doesnt_exist(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:user-worker-run-create"),
                data={"worker_version_id": str(self.other_version.id), "configuration_id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"configuration_id": ['Invalid pk "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa" - object does not exist.']})

    @patch("arkindex.users.managers.BaseACLManager.filter_rights", return_value=ModelVersion.objects.none())
    def test_create_user_run_model_no_access(self, filter_rights_mock):
        self.client.force_login(self.user)

        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:user-worker-run-create"),
                data={"worker_version_id": str(self.other_version.id), "model_version_id": str(self.model_version.id)}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {"model_version_id": ["You do not have guest access to this model."]})

        self.assertListEqual(filter_rights_mock.call_args_list, [
            call(self.user, Model, Role.Guest.value),
            call(self.user, Model, Role.Contributor.value),
        ])

    def test_create_user_run_model_archived(self):
        self.model.archived = datetime.now(timezone.utc)
        self.model.save()
        self.model.memberships.create(user=self.user, level=Role.Contributor.value)
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:user-worker-run-create"),
                data={"worker_version_id": str(self.other_version.id), "model_version_id": str(self.model_version.id)}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {"model_version_id": ["The model used to create a local worker run must not be archived."]})

    def test_create_user_run_full(self):
        self.model_version.tag = "0.1.3"
        self.model_version.state = ModelVersionState.Available
        self.model_version.save()
        self.model.memberships.create(user=self.user, level=Role.Guest.value)
        test_configuration = WorkerConfiguration.objects.create(worker=self.other_version.worker, name="Some configuration", configuration={"param": "value"})

        self.client.force_login(self.user)

        with self.assertNumQueries(10):
            response = self.client.post(
                reverse("api:user-worker-run-create"),
                data={
                    "worker_version_id": str(self.other_version.id),
                    "model_version_id": str(self.model_version.id),
                    "configuration_id": str(test_configuration.id)
                }
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        created_run = WorkerRun.objects.get(version=self.other_version)
        self.assertDictEqual(response.json(), {
            "id": str(created_run.id),
            "parents": [],
            "configuration": {
                "archived": False,
                "configuration": {"param": "value"},
                "id": str(test_configuration.id),
                "name": "Some configuration"
            },
            "model_version": {
                "configuration": {},
                "id": str(self.model_version.id),
                "model": {
                    "id": str(self.model.id),
                    "name": "Some model"
                },
                "size": 8,
                "state": "available",
                "tag": "0.1.3"
            },
            "worker_version": {
                "configuration": {
                    "name": "bartimaeus",
                    "category": "level 14 djinni"
                },
                "created": self.other_version.created.isoformat().replace("+00:00", "Z"),
                "docker_image_iid": None,
                "gpu_usage": "disabled",
                "id": str(self.other_version.id),
                "model_usage": FeatureUsage.Disabled.value,
                "revision": None,
                "state": "created",
                "version": 2,
                "worker": {
                    "id": str(self.other_version.worker.id),
                    "name": "Custom worker",
                    "slug": "custom",
                    "type": "custom"
                }
            },
            "summary": f"Worker Custom worker @ version 2 with model Some model @ {str(self.model_version.id)[:6]} using configuration 'Some configuration'",
            "process": {
                "activity_state": "disabled",
                "corpus": None,
                "chunks": 1,
                "id": str(self.local_process.id),
                "mode": "local",
                "name": None,
                "state": "unscheduled",
                "use_cache": False,
            },
            "use_gpu": False
        })

    def test_create_user_run_duplicate(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:user-worker-run-create"),
                data={
                    "worker_version_id": str(self.custom_version.id)
                }
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "detail": ["A local worker run with the same worker version, model version and configuration already exists."],
            "id": [str(self.local_run.id)]
        })
