import logging

from django.conf import settings
from django.contrib.auth import login, logout
from django.contrib.auth.forms import PasswordResetForm
from django.contrib.auth.tokens import default_token_generator
from django_rq.queues import get_queue
from django_rq.settings import QUEUES
from drf_spectacular.types import OpenApiTypes
from drf_spectacular.utils import OpenApiParameter, extend_schema, extend_schema_view
from rest_framework import status
from rest_framework.exceptions import AuthenticationFailed, NotFound, PermissionDenied, ValidationError
from rest_framework.generics import CreateAPIView, ListAPIView, RetrieveDestroyAPIView, RetrieveUpdateDestroyAPIView
from rest_framework.response import Response
from rq.job import JobStatus

from arkindex.project.permissions import IsAuthenticated, IsAuthenticatedOrReadOnly, IsVerified
from arkindex.project.triggers import send_verification_email
from arkindex.users.models import User
from arkindex.users.serializers import (
    EmailLoginSerializer,
    JobSerializer,
    NewUserSerializer,
    PasswordResetConfirmSerializer,
    PasswordResetSerializer,
    UserEmailVerificationSerializer,
    UserSerializer,
)

logger = logging.getLogger(__name__)

# Process tasks running in RQ are hidden from user jobs
VISIBLE_QUEUES = [
    q for q in QUEUES.keys()
    if q != "tasks"
]


@extend_schema(tags=["users"])
@extend_schema_view(
    get=extend_schema(description="Retrieve information about the authenticated user"),
    patch=extend_schema(description="Update a user's password. This action is not allowed to users without confirmed e-mails."),
    put=extend_schema(description="Update a user's password. This action is not allowed to users without confirmed e-mails."),
    delete=extend_schema(operation_id="Logout", description="Log out from the API")
)
class UserRetrieve(RetrieveUpdateDestroyAPIView):
    """
    Retrieve information about the authenticated user
    """
    permission_classes = (IsAuthenticatedOrReadOnly, )
    serializer_class = UserSerializer

    def get_object(self):
        return self.request.user

    def retrieve(self, *args, **kwargs):
        if not self.request.user.is_authenticated:
            # Include features on GET to let the frontend still know the instance features when not logged in
            return Response(
                {
                    "detail": PermissionDenied.default_detail,
                    "features": settings.ARKINDEX_FEATURES,
                },
                status=status.HTTP_403_FORBIDDEN,
            )

        return super().retrieve(*args, **kwargs)

    def perform_destroy(self, instance):
        logout(self.request)


@extend_schema_view(post=extend_schema(
    operation_id="Register",
    responses={201: UserSerializer},
    tags=["users"],
))
class UserCreate(CreateAPIView):
    """
    Register as a new user.
    """
    serializer_class = NewUserSerializer

    def create(self, request, *args, **kwargs):
        if not settings.ARKINDEX_FEATURES["signup"]:
            raise ValidationError(["Registration has been disabled on this instance."])

        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        user = serializer.save()
        login(self.request, user)

        send_verification_email(user)

        return Response(UserSerializer(user).data, status=status.HTTP_201_CREATED)


@extend_schema_view(post=extend_schema(
    operation_id="Login",
    responses={201: UserSerializer},
    tags=["users"]
))
class UserEmailLogin(CreateAPIView):
    """
    Login using an email and a password
    """
    serializer_class = EmailLoginSerializer

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        try:
            user = User.objects.get_by_natural_key(serializer.validated_data["email"])
        except User.DoesNotExist:
            raise AuthenticationFailed

        if not user.has_usable_password() or not user.check_password(serializer.validated_data["password"]):
            raise AuthenticationFailed

        login(self.request, user)

        headers = self.get_success_headers(serializer.data)
        return Response(
            UserSerializer(user).data,
            status=status.HTTP_201_CREATED,
            headers=headers,
        )


@extend_schema_view(post=extend_schema(
    operation_id="VerifyEmail",
    tags=["users"],
))
class UserEmailVerification(CreateAPIView):
    """
    Verify a user's email address
    """
    serializer_class = UserEmailVerificationSerializer


@extend_schema_view(post=extend_schema(operation_id="ResetPassword", tags=["users"]))
class PasswordReset(CreateAPIView):
    """
    Start a password reset flow
    """
    serializer_class = PasswordResetSerializer

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        form = PasswordResetForm(serializer.validated_data)

        if form.is_valid():
            form.save(
                request=self.request,
                use_https=self.request.is_secure(),
                token_generator=default_token_generator,
                email_template_name="registration/password_reset_email.html",
                subject_template_name="registration/password_reset_subject.txt",
            )
            headers = self.get_success_headers(serializer.data)
            return Response(
                status=status.HTTP_201_CREATED,
                headers=headers,
            )
        return Response(
            form.errors.get_json_data(),
            status=status.HTTP_400_BAD_REQUEST,
            headers=self.default_response_headers
        )


@extend_schema_view(post=extend_schema(operation_id="PasswordResetConfirm", tags=["users"]))
class PasswordResetConfirm(CreateAPIView):
    """
    Confirm a password reset using data from the confirmation e-mail
    """
    serializer_class = PasswordResetConfirmSerializer


@extend_schema(tags=["jobs"])
class JobList(ListAPIView):
    """
    List asynchronous jobs linked to the current user.
    """
    permission_classes = (IsVerified, )
    serializer_class = JobSerializer
    pagination_class = None

    def get_queryset(self):
        if not self.request:
            return []
        return self.request.user.get_rq_jobs()


@extend_schema(
    tags=["jobs"],
    parameters=[
        OpenApiParameter(
            "id",
            location=OpenApiParameter.PATH,
            type=OpenApiTypes.UUID,
        )
    ]
)
class JobRetrieve(RetrieveDestroyAPIView):
    """
    Retrieve a single job by ID.
    """
    permission_classes = (IsVerified, )
    serializer_class = JobSerializer

    def get_object(self):
        for queue_name in VISIBLE_QUEUES:
            job = get_queue(queue_name).fetch_job(str(self.kwargs["pk"]))
            if not job:
                continue
            if job.user_id != self.request.user.id:
                raise NotFound
            return job
        raise NotFound

    def perform_destroy(self, instance):
        if instance.get_status(refresh=False) == JobStatus.STARTED:
            raise ValidationError(["Cannot delete a running job."])
        instance.delete()


@extend_schema_view(post=extend_schema(
    operation_id="SendVerificationEmail",
    request=None,
    responses={202: None},
    tags=["users"],
))
class SendVerificationEmail(CreateAPIView):
    """
    Send a new validation email in case the user is not verified
    """
    permission_classes = (IsAuthenticated, )

    def create(self, request, *args, **kwargs):
        if self.request.user.verified_email:
            raise ValidationError(["Your account is already verified"])
        send_verification_email(self.request.user)
        return Response(status=status.HTTP_202_ACCEPTED)
